from helpers.data_processing_helpers import (get_all_data, combine_data, split_data)
from helpers.model_helpers import (get_type_dfs, get_models, cross_val_models,
                                    load_best_models, find_df, append_pred, get_date_indexes,
                                    cross_val_monolith)
from helpers.visualization_helpers import (get_values_from_csv, map_pred_to_dates, plot_daily_predictions)
from joblib import load
import pandas as pd

def init_data():
    data = get_all_data()
    combined_data = combine_data(data)
    types = combined_data.ResourceType.unique()
    df_types = get_type_dfs(types, combined_data.columns)
    df_types = split_data(df_types, combined_data, types)
    return data, combined_data, df_types, types

def demo_multi_model():
    data, combined_data, df_types, types = init_data()
    models = get_models()
    cross_val_models(df_types=df_types, models=models, types=types, n_splits=5, tune=False)
    get_values_from_csv(filename='results/cross_validation.csv')

def demo_multi_model_plot():
    data, combined_data, df_types, types = init_data()
    loaded_models, properties = load_best_models('results/cross_validation.csv_best_models.csv')
    date_mapping = get_date_indexes(data)
    for i in range(len(loaded_models)):
        model = loaded_models[i]
        df = find_df(properties[i][0], df_types)
        pred_data = append_pred(model, df)
        plot_data = map_pred_to_dates(pred_data, date_mapping)
        plot_daily_predictions(plot_data, properties[i], total=False)

def demo_monolith():
    data, combined_data, df_types, types = init_data()
    models = get_models()
    cross_val_monolith(data=combined_data, model=models[0], tune=False)
    print('yes')

def demo_monolith_plot():
    data, combined_data, df_types, types = init_data()
    model = load('models/cross-val/monlith.joblib')
    date_mapping = get_date_indexes(data)
    pred_data = append_pred(model=model, data=combined_data)
    plot_data = map_pred_to_dates(pred_data, date_mapping)
    plot_daily_predictions(plot_data, None, monolith=True, total=False)