from datetime import date, timedelta
import matplotlib.pyplot as plt
import csv
from operator import itemgetter
import pandas as pd
from sklearn.metrics import mean_squared_error, r2_score

# These functions are for plotting and visualizing data.

# General function for plotting cost and grouping data based on some parameter.
# Param: field in data, example: ResourceType
# Goal: value of param to be searched for, example: microsoft.storage/storageaccounts
def print_cost(data, param, goal, avg=False, rolling=False, save_img=False, drift_points=None):
    costs = get_cost_arr(data, param, goal, avg, rolling)
    dates = get_dates()
    figure = plt.figure()
    figure.set_figwidth(10)
    plt.plot(dates, costs)
    plt.title('Total cost', fontsize=15) if param == 'none' else plt.title(f'{goal}')
    plt.xlabel('Date', fontsize=13)
    plt.ylabel('Price (USD)', fontsize=13)
    plt.title('Total cost') if param == 'none' else plt.title(f'{goal}')
    plt.xlabel('Date')
    plt.ylabel('Price (USD)')
    
    if drift_points:
        for drift_point in drift_points:
            plt.axvline(x=dates[drift_point], color='r', linestyle='--')
    
    if save_img:
        safe_goal = goal.replace("/", "_")
        if param == 'SubscriptionId':
            filename = 'images/Subcription.png'
            plt.title('Subscription')
        else:
            filename = f"images/{safe_goal}_Cost_Trend.png"
        plt.savefig(filename)
    plt.show()

# Helper function for print_cost, returns an array with the costs of each day.
def get_cost_arr(data, param, goal, avg, rolling):
    costs = []
    for i in range(365):
        cost = 0
        resources = 0
        for j in range(len(data[i])):
            if param == 'none':
                cost += float(data[i]['CostUSD'][j])
            elif param == 'SubscriptionId':
                if goal in data[i]['ResourceId'][j]:
                    cost += float(data[i]['CostUSD'][j])
                    resources += 1
                    continue
            elif data[i][param][j] == goal:
                cost += float(data[i]['CostUSD'][j])
                resources += 1
        if avg:
            costs.append(cost / resources) if resources != 0 else costs.append(0)   
        else:
            costs.append(cost)
    
    if rolling:
        costs_series = pd.Series(costs)
        window_size = 30
        rolling_mean = costs_series.rolling(window=window_size).mean()
        return rolling_mean.tolist()
    return costs

# Returns an array of dates in the desired format.
def get_dates():
    dates = []
    start_date = date(2022, 1, 1) 
    end_date = date(2022, 12, 31)    

    delta = end_date - start_date   

    for i in range(delta.days + 1):
        day = start_date + timedelta(days=i)
        dates.append(day)
    return dates


def get_cost_type_arr(data, tp):
    type_names = []
    type_cost = []
    
    if not tp in data[0]:
        print('Type not found!')
        return
    
    for i in range(365):
        for j in range(len(data[i])):
            if data[i][tp][j] in type_names:
                index = type_names.index(data[i][tp][j])
                type_cost[index] += float(data[i]['Cost'][j])
            else:
                type_names.append(data[i][tp][j])
                type_cost.append(float(data[i]['Cost'][j]))
            
    return type_names, type_cost

# Fjern?
def plot_cost_type_arr(data, tp):
    type_names, type_cost = get_cost_type_arr(data, tp)
    
    # sort the lists in descending order by cost
    sorted_idx = sorted(range(len(type_cost)), key=lambda k: -type_cost[k])
    type_names = [type_names[i] for i in sorted_idx]
    type_cost = [type_cost[i] for i in sorted_idx]


    plt.xlabel(tp)
    plt.ylabel('Cost')
    plt.bar(type_names, type_cost)
    plt.xticks(rotation=300, ha="left")
    plt.show()

# Fjern?
def plot_true_vs_predicted(y_true, y_pred, model_name):
    plt.scatter(y_true, y_pred, alpha=0.3, c='crimson', label='Predicted')
    plt.xscale('log')
    plt.yscale('log')
    plt.xlabel('True Values')
    plt.ylabel('Predicted Values')
    plt.title(f'True vs Predicted Values ({model_name})')


    # Add a reference line representing a perfect match between true and predicted values
    min_val = min(min(y_true), min(y_pred))
    max_val = max(max(y_true), max(y_pred))
    plt.plot([min_val, max_val], [min_val, max_val], color='red', label='Ideal')

    plt.legend()
    plt.show()

# Defines dictionaries for short form of the resource types and model names.
# Returns a shortened for name visualization based on the two.
def get_name(model, resource_type):
    types = {
        'microsoft.storage/storageaccounts': 'Storage',
        'microsoft.compute/disks': 'Disks',
        'microsoft.compute/virtualmachines': 'VMs',
        'microsoft.recoveryservices/vaults': 'Recovery valuts',
        'microsoft.compute/virtualmachinescalesets': 'VM scls',
        'microsoft.keyvault/vaults': 'Key vaults',
        'microsoft.network/loadbalancers': 'Load balancers',
        'microsoft.network/publicipaddresses': 'Public IP',
        'microsoft.operationalinsights/workspaces': 'Workspaces',
        'microsoft.network/bastionhosts': 'Bastion',
        'microsoft.network/privatednszones': 'Private DNS',
        'microsoft.network/virtualnetworkgateways': 'VM network',
        'microsoft.network/networkwatchers': 'Network watchers',
        'microsoft.compute/galleries': 'Galleries',
        'microsoft.servicebus/namespaces': 'Namespaces',
        'microsoft.security/pricings': 'Pricings',
        'microsoft.containerinstance/containergroups': 'Containers',
        'microsoft.network/azurefirewalls': 'Firewalls',
        'microsoft.network/connections': 'Connections',
        'microsoft.containerregistry/registries': 'Registries',
        'microsoft.insights/actiongroups': 'Action groups',
        'microsoft.network/expressroutecircuits': 'Express Routes',
        'microsoft.insights/metricalerts': 'Metric alerts',
        'microsoft.compute/snapshots': 'Snapshots',
        'microsoft.compute/images': 'Images',
        'microsoft.compute/restorepointcollections': 'Restore points',
        'microsoft.network/privateendpoints': 'Private endpoints',
        'microsoft.insights/scheduledqueryrules': 'Scheduled query rules'
    }

    models = {
        'RandomForestRegressor': 'RF',
        'KNeighborsRegressor': 'KNN',
        'MLPRegressor': 'MLP',
        'Ridge': 'Ridge',
        'GradientBoostingRegressor': 'GBR'
    }

    name = f'{types[resource_type]}, {models[model]}'

    return name

# Converts a list of strings to a list of floats.
def convert_strings(list):
    list = [float(x) for x in list]
    
    return list

# Returns the best values of each resource type from a csv file of results based on r2.
def get_best_values(list):
    best_values = []
    completed = []

    list = sorted(list, key=itemgetter(3), reverse=True)

    for i in range(1, len(list)):
        if list[i][0] not in completed:
            best_values.append(list[i])
            completed.append(list[i][0])

    return best_values

def get_values_from_csv(filename, plot=False, hyperparameters=False):
    file = open(filename)
    csvreader = csv.reader(file)

    header = []
    header = next(csvreader)

    rows = []
    for row in csvreader:
        rows.append(row)

    rows = save_best(rows, filename, hyperparameters)

    if plot:

        mses = []
        r2s = []
        names = []

        for row in rows:
            if float(row[2]) < 10 and float(row[3]) > 0.3:
                mses.append(row[2])
                r2s.append(row[3])
                names.append(get_name(resource_type=row[0], model=row[1]))

        
        return mses, r2s, names
    
    else:
        return rows

def generate_performance_plot():
    mses = []
    r2s = []
    names = []

    mses, r2s, names = get_values_from_csv('results.csv', plot=True)

    mses = convert_strings(mses)
    r2s = convert_strings(r2s)

    fig = plt.figure(figsize=(16, 10), dpi= 80, facecolor='w', edgecolor='k')
    colors = [plt.cm.tab10(i/float(len(names)-1)) for i in range(len(names))]
    for i, name in enumerate(names):
        plt.scatter(r2s[i], mses[i], s=200, c=colors[i], label=name, edgecolors='black', linewidths=.5)
    plt.gca().set(xlim=(0.2, 1.0), ylim=(-0.5, 5),
                xlabel='R2', ylabel='MSE')

    plt.xticks(fontsize=12); plt.yticks(fontsize=12)
    plt.title("R2 and MSE performance", fontsize=22)
    plt.legend(fontsize=12)     
    plt.savefig('foo2.png')

# Writes the performance of the best models to a seperate file.
def save_best(list, filename, hyperparameters):
    list = get_best_values(list)
    file = f"{filename}_best_models.csv"
    
    f = open(file, "a")
    if hyperparameters:
        f.write("ResourceType,Model,MSE,R2,Diff%,Hyperparameters\n")
    else: 
        f.write("ResourceType,Model,MSE,R2,Diff%\n")
    f.close()
    for i in list:
        f = open(file, "a")
        if hyperparameters:
            f.write(f"{i[0]},{i[1]},{i[2]},{i[3]},{i[4]},{i[5]}\n")
        else:
            f.write(f"{i[0]},{i[1]},{i[2]},{i[3]},{i[4]}\n")
        f.close()
    
    return list


def plot_pred(model_name, data, date_index_df):
    pred = data['pred']
    y = data['CostUSD']
    X = map_pred_to_dates(df=data, date_indexes=date_index_df)
    figure = plt.figure()
    figure.set_figwidth(10)
    plt.plot(X, y)
    plt.plot(X, pred)
    plt.title(f'Total cost, {model_name}')
    plt.xlabel('Date')
    plt.ylabel('Price (NOK)')
    plt.show()

def map_pred_to_dates(df, mapping):
    dates = [pd.DataFrame(columns=df.columns) for _ in range(365)]
    for _, row in df.iterrows():
        old_index = row.name
        df_index = mapping[old_index]
        dates[df_index] = dates[df_index].append(row)

    return dates

def plot_daily_predictions(dataframes_list, properties, monolith=False, total=False):
    name = 'monolith' if monolith else get_name(properties[1], properties[0])
    real_values_sum = []
    predicted_values_sum = []
    performance = []
    days = list(range(len(dataframes_list)))

    for df in dataframes_list:
        real_values_sum.append(df['CostUSD'].sum())
        predicted_values_sum.append(df['pred'].sum())

    avg_real = get_avg(real_values_sum)
    avg_pred = get_avg(predicted_values_sum)

    mse = mean_squared_error(real_values_sum, predicted_values_sum)
    r2 = r2_score(real_values_sum, predicted_values_sum)
    diff = abs(avg_pred - avg_real / avg_real) * 100

    performance.append(mse)
    performance.append(r2)
    performance.append(diff)

    if total:
        return real_values_sum, predicted_values_sum, performance

    plt.figure(figsize=(10, 6))
    plt.plot(days, real_values_sum, alpha=0.7, label='Real Values', marker='o', linestyle='-', linewidth=2)
    plt.plot(days, predicted_values_sum, alpha=0.7, label='Predicted Values', marker='x', linestyle='--', linewidth=2)
    plt.xlabel('Day', fontsize=13)
    plt.ylabel('Sum of Prices (USD)', fontsize=13)
    plt.title(f'Real vs. Predicted Sum of Prices per Day \n {name}', fontsize=15)
    plt.legend()
    plt.savefig(f'images/{name}')

def get_avg(list):
    sum = 0
    for i in list:
        sum+=i
    return sum/len(list)

def format_hyperparams(filename):
    input_file = filename
    output_file = f'formatted_{filename}'

    with open(input_file, 'r') as f_in, open(output_file, 'w') as f_out:
        # Copy the header row
        f_out.write(next(f_in))
    
    for line in f_in:
        dict_start = line.index('{')
        new_line = line[:dict_start] + '"' + line[dict_start:].strip() + '"\n'
        f_out.write(new_line)